#include <string.h>

#include <glib/gutils.h>
#include <libgnome/gnome-i18n.h>
#include <libgnomeui/gnome-stock-icons.h>

#include <gedit/gedit-menus.h>
#include <gedit/gedit-plugin.h>
#include <gedit/gedit-debug.h>
#include <gedit/gedit-prefs-manager.h>

#define MENU_COMPLETE_NEXT_LABEL  N_("Complete Next")
#define MENU_COMPLETE_NEXT_NAME   "Complete Next"
#define MENU_COMPLETE_NEXT_TIP    N_("Complete the current word by finding a similar word later in the document")

#define MENU_COMPLETE_PREV_LABEL  N_("Complete Previous")
#define MENU_COMPLETE_PREV_NAME   "Complete Previous"
#define MENU_COMPLETE_PREV_TIP    N_("Complete the current word by finding a similar word earlier in the document")

#define MENU_ITEM_PATH	 	  "/menu/Edit/EditOps_5/"

G_MODULE_EXPORT GeditPluginState update_ui(GeditPlugin *plugin, BonoboWindow *window);
G_MODULE_EXPORT GeditPluginState activate(GeditPlugin *pd);
G_MODULE_EXPORT GeditPluginState deactivate(GeditPlugin *pd);
G_MODULE_EXPORT GeditPluginState init(GeditPlugin *pd);

gboolean is_seperator_pred(gunichar ch, gpointer user_data) {
	return g_unichar_isspace(ch) || g_unichar_ispunct(ch);
}
gboolean is_not_seperator_pred(gunichar ch, gpointer user_data) {
	return !is_seperator_pred(ch, user_data);
}

static void complete_next_cb(BonoboUIComponent *uic, gpointer user_data, const gchar *verbname) {
	GeditView     *view;
	GeditDocument *doc;

	gedit_debug(DEBUG_PLUGINS, "");

	view = gedit_get_active_view();
	g_return_if_fail(view != NULL);

	doc = gedit_view_get_document(view);
	g_return_if_fail(doc != NULL);

	gedit_document_begin_user_action(doc);
	{
		GtkTextIter cur, begin, end;
		gchar *prefix = NULL;

		gtk_text_buffer_get_iter_at_mark(GTK_TEXT_BUFFER(doc), &cur, gtk_text_buffer_get_insert(GTK_TEXT_BUFFER(doc)));
		gtk_text_buffer_get_start_iter(GTK_TEXT_BUFFER(doc), &begin);
		gtk_text_buffer_get_end_iter(GTK_TEXT_BUFFER(doc), &end);

		{ /* Find out what the prefix we're matching is (the current word) */
			GtkTextIter word_begin = cur;

			if(gtk_text_iter_backward_find_char(&word_begin, &is_seperator_pred, NULL, NULL)) {
				if(gtk_text_iter_forward_char(&word_begin))
					prefix = gtk_text_iter_get_text(&word_begin, &cur);
			} else
				prefix = gtk_text_iter_get_text(&begin, &cur);
		}

		if(prefix == NULL) {
			/* No selected prefix, just find the next word */
			GtkTextIter next_begin = cur;

			if(gtk_text_iter_forward_find_char(&next_begin, &is_not_seperator_pred, NULL, NULL)) {
				GtkTextIter next_end = next_begin;

				if(gtk_text_iter_forward_find_char(&next_end, &is_seperator_pred, NULL, NULL))
					gtk_text_iter_backward_char(&next_begin);
				else
					next_end = end;

				gtk_text_buffer_insert_range(GTK_TEXT_BUFFER(doc), &cur, &next_begin, &next_end);
			}

		} else {
			/* Find the next word that has this prefix */
			GtkTextIter word_end = cur;
			GtkTextIter match_begin, match_end = word_end;

			while(gtk_text_iter_forward_search(&match_end, prefix, 0U, &match_begin, &match_end, NULL)) {
				/* If match_end is a seperator, the prefix was an exact match, try again */
				if(is_seperator_pred(gtk_text_iter_get_char(&match_end), NULL)) {
					continue;
				}
				{
					GtkTextIter suffix_begin = match_end, suffix_end = match_end;

					gtk_text_iter_forward_find_char(&suffix_end, &is_seperator_pred, NULL, NULL);

					{
						/* We found a new suffix, use it */
						gtk_text_buffer_insert_range(GTK_TEXT_BUFFER(doc), &cur, &suffix_begin, &suffix_end);

						/* All done */
						break;
					}
				}
			}
		}

		g_free(prefix);
	}
	gedit_document_end_user_action(doc);
}

static void complete_prev_cb(BonoboUIComponent *uic, gpointer user_data, const gchar *verbname) {
	GeditView     *view;
	GeditDocument *doc;

	gedit_debug(DEBUG_PLUGINS, "");

	view = gedit_get_active_view();
	g_return_if_fail(view != NULL);

	doc = gedit_view_get_document(view);
	g_return_if_fail(doc != NULL);

	gedit_document_begin_user_action(doc);
	{
		GtkTextIter cur, begin, end, word_begin;
		gchar *prefix = NULL;

		gtk_text_buffer_get_iter_at_mark(GTK_TEXT_BUFFER(doc), &cur, gtk_text_buffer_get_insert(GTK_TEXT_BUFFER(doc)));
		gtk_text_buffer_get_start_iter(GTK_TEXT_BUFFER(doc), &begin);
		gtk_text_buffer_get_end_iter(GTK_TEXT_BUFFER(doc), &end);

		{ /* Find out what the prefix we're matching is (the current word) */
			word_begin = cur;

			if(gtk_text_iter_backward_find_char(&word_begin, &is_seperator_pred, NULL, NULL)) {
				if(gtk_text_iter_forward_char(&word_begin))
					prefix = gtk_text_iter_get_text(&word_begin, &cur);
			}
		}

		if(prefix == NULL) {
			/* No selected prefix, just find the previous word */
			GtkTextIter prev_end = cur;

			if(gtk_text_iter_backward_find_char(&prev_end, &is_not_seperator_pred, NULL, NULL)) {
				GtkTextIter prev_begin = prev_end;

				gtk_text_iter_forward_char(&prev_end);

				if(gtk_text_iter_backward_find_char(&prev_begin, &is_seperator_pred, NULL, NULL))
					gtk_text_iter_forward_char(&prev_begin);
				else
					prev_begin = begin;

				gtk_text_buffer_insert_range(GTK_TEXT_BUFFER(doc), &cur, &prev_begin, &prev_end);
			}
		} else {
			/* Find the previous word that has this prefix */
			GtkTextIter match_begin = word_begin, match_end;

			while(gtk_text_iter_backward_search(&match_begin, prefix, 0U, &match_begin, &match_end, NULL)) {
				/* Check if the match is really at the beginning of a word */
				if(!gtk_text_iter_equal(&begin, &match_begin)) {
					GtkTextIter match_word_begin = match_begin;

					if(gtk_text_iter_backward_find_char(&match_word_begin, &is_seperator_pred, NULL, NULL)) {
						if(!gtk_text_iter_forward_char(&match_word_begin) || !gtk_text_iter_equal(&match_word_begin, &match_begin)) {
							/* Not the beginning of a word, try again */
							continue;
						}
					} else {
						/* Not the beginning of a word, try again */
						continue;
					}
				}
				/* If match_end is a seperator, the prefix was an exact match, try again */
				if(is_seperator_pred(gtk_text_iter_get_char(&match_end), NULL)) {
					continue;
				}
				{
					GtkTextIter suffix_begin = match_end, suffix_end = match_end;

					if(gtk_text_iter_forward_find_char(&suffix_end, &is_seperator_pred, NULL, NULL)) {
						/* We found a new suffix, use it */
						gtk_text_buffer_insert_range(GTK_TEXT_BUFFER(doc), &cur, &suffix_begin, &suffix_end);

						/* All done */
						break;
					}
				}
			}
		}

		g_free(prefix);
	}
	gedit_document_end_user_action(doc);
}

G_MODULE_EXPORT GeditPluginState update_ui(GeditPlugin *plugin, BonoboWindow *window) {
	BonoboUIComponent *uic;
	GeditDocument     *doc;
	
	gedit_debug(DEBUG_PLUGINS, "");
	
	g_return_val_if_fail(window != NULL, PLUGIN_ERROR);

	uic = gedit_get_ui_component_from_window(window);

	doc = gedit_get_active_document();

	if((doc == NULL) || (gedit_document_is_readonly(doc))) {		
		gedit_menus_set_verb_sensitive(uic, "/commands/" MENU_COMPLETE_NEXT_NAME, FALSE);
		gedit_menus_set_verb_sensitive(uic, "/commands/" MENU_COMPLETE_PREV_NAME, FALSE);
	} else {
		gedit_menus_set_verb_sensitive(uic, "/commands/" MENU_COMPLETE_NEXT_NAME, TRUE);
		gedit_menus_set_verb_sensitive(uic, "/commands/" MENU_COMPLETE_PREV_NAME, TRUE);
	}

	return PLUGIN_OK;
}

G_MODULE_EXPORT GeditPluginState activate(GeditPlugin *pd) {
	GList             *top_windows;
	BonoboUIComponent *ui_component;

        gedit_debug(DEBUG_PLUGINS, "");

        top_windows = gedit_get_top_windows();
        g_return_val_if_fail(top_windows != NULL, PLUGIN_ERROR);

        while(top_windows) {
		gedit_menus_add_menu_item(BONOBO_WINDOW(top_windows->data),
			MENU_ITEM_PATH, MENU_COMPLETE_NEXT_NAME,
		        MENU_COMPLETE_NEXT_LABEL, MENU_COMPLETE_NEXT_TIP, GTK_STOCK_GO_FORWARD,
		        &complete_next_cb);

		gedit_menus_add_menu_item(BONOBO_WINDOW(top_windows->data),
			MENU_ITEM_PATH, MENU_COMPLETE_PREV_NAME,
			MENU_COMPLETE_PREV_LABEL, MENU_COMPLETE_PREV_TIP, GTK_STOCK_GO_BACK,
			&complete_prev_cb);

		ui_component = gedit_get_ui_component_from_window(BONOBO_WINDOW(top_windows->data));

		bonobo_ui_component_set_prop(ui_component, "/commands/" MENU_COMPLETE_NEXT_NAME, "accel", "*Control*bracketright", NULL);
		bonobo_ui_component_set_prop(ui_component, "/commands/" MENU_COMPLETE_PREV_NAME, "accel", "*Control*bracketleft", NULL);

                pd->update_ui(pd, BONOBO_WINDOW(top_windows->data));

                top_windows = g_list_next(top_windows);
        }

        return PLUGIN_OK;
}

G_MODULE_EXPORT GeditPluginState deactivate(GeditPlugin *pd) {
	gedit_menus_remove_menu_item_all(MENU_ITEM_PATH, MENU_COMPLETE_NEXT_NAME);
	gedit_menus_remove_menu_item_all(MENU_ITEM_PATH, MENU_COMPLETE_PREV_NAME);

	return PLUGIN_OK;
}

G_MODULE_EXPORT GeditPluginState init(GeditPlugin *pd) {
	/* initialize */
	gedit_debug(DEBUG_PLUGINS, "");
     
	pd->private_data = NULL;
		
	return PLUGIN_OK;
}
